/*
 *  Copyright (C) 2014 Andreas Steinmetz <ast@domdv.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <linux/icmp.h>
#include <netinet/icmp6.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <signal.h>
#include <time.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <poll.h>
#include <stdio.h>

#undef _POSIX_SOURCE
#define _LINUX_FS_H
#include <sys/syscall.h>
#include <linux/capability.h>
#include <sys/prctl.h>

extern int capget(cap_user_header_t header, cap_user_data_t data);
extern int capset(cap_user_header_t header, const cap_user_data_t data);

static int drop(void)
{
	struct __user_cap_header_struct h;
	struct __user_cap_data_struct c[2];

	memset(&h,0,sizeof(h));
	h.version=_LINUX_CAPABILITY_VERSION_2;

	memset(&c,0,sizeof(c));
	capget(&h,c);
	if(!(c[0].effective&(1<<CAP_NET_RAW))||
		!(c[0].effective&(1<<CAP_NET_RAW)))return -1;

	if(!geteuid())
	{
		memset(&h,0,sizeof(h));
		h.version=_LINUX_CAPABILITY_VERSION_2;
		memset(&c,0,sizeof(c));
		c[0].effective=1<<CAP_NET_RAW|1<<CAP_SETUID;
		c[0].permitted=1<<CAP_NET_RAW|1<<CAP_SETUID;
		capset(&h,c);
		prctl(PR_SET_KEEPCAPS,1,0,0,0);
		if(seteuid(getuid()))return -1;
		memset(&h,0,sizeof(h));
		h.version=_LINUX_CAPABILITY_VERSION;
		memset(&c,0,sizeof(c));
		c[0].effective=1<<CAP_NET_RAW;
		c[0].permitted=1<<CAP_NET_RAW;
		capset(&h,c);
		prctl(PR_SET_KEEPCAPS,0,0,0,0);
	}

	return 0;
}

static void dropall(void)
{
	struct __user_cap_header_struct h;
	struct __user_cap_data_struct c[2];

	memset(&h,0,sizeof(h));
	h.version=_LINUX_CAPABILITY_VERSION_2;
	h.pid=0;
	memset(&c,0,sizeof(c));
	capset(&h,c);
}

static int sendtcp(int s,int family,int port,int *srcsock,int *srcport,
	unsigned int sum)
{
	socklen_t l;
	unsigned char *data;
	struct tcphdr th;
	struct sockaddr_storage a;

	if((*srcsock=socket(family,SOCK_STREAM,0))==-1)
	{
		perror("socket");
		return -1;
	}

	memset(&a,0,sizeof(a));
	a.ss_family=family;

	if(bind(*srcsock,(struct sockaddr *)&a,sizeof(a)))
	{
		perror("bind");
		close(*srcsock);
		return -1;
	}

	memset(&a,0,sizeof(a));
	l=sizeof(a);
	if(getsockname(*srcsock,(struct sockaddr *)&a,&l))
	{
		perror("getsockname");
		close(*srcsock);
		return -1;
	}

	if(family==AF_INET)*srcport=ntohs(((struct sockaddr_in *)&a)->sin_port);
	else *srcport=ntohs(((struct sockaddr_in6 *)&a)->sin6_port);

	memset(&th,0,sizeof(th));
	th.source=htons(*srcport);
	th.dest=htons(port);
	th.seq=htonl(time(NULL));
	th.doff=5;
	th.syn=1;
	th.window=htonl(6000);

	for(data=(unsigned char *)&th,l=0;l<sizeof(struct tcphdr);l+=2)
	{
		sum+=data[l];
		sum+=((unsigned int)data[l+1])<<8;
	}

	while(sum&0xffff0000)sum=(sum>>16)+(sum&0xffff);
	sum=~sum;

	data=(unsigned char *)&th.check;
	data[0]=sum&0xff;
	data[1]=(unsigned char)(sum>>8);

	if(send(s,&th,sizeof(th),MSG_DONTWAIT)!=sizeof(th))
	{
		perror("send");
		close(*srcsock);
		return -1;
	}

	return 0;
}

static int recvtcp(int s,int family,int port,int srcport)
{
	int l;
	struct tcphdr *th;
	unsigned char msg[128];

	if((l=recv(s,&msg,sizeof(msg),0))==-1)
	{
		perror("recv");
		return -1;
	}

	if(family==AF_INET)
	{
		if(l<sizeof(struct iphdr))return -1;
		l-=sizeof(struct iphdr);
		th=(struct tcphdr *)(msg+sizeof(struct iphdr));
	}
	else th=(struct tcphdr *)(msg);

	if(l<sizeof(struct tcphdr))return -1;
	if(th->source!=htons(port)||th->dest!=htons(srcport))return -1;
	if(!th->syn||!th->ack)return 1;

	return 0;
}

static int socktcp(int family,unsigned char *addr,unsigned int *sum)
{
	int s;
	int i;
	union
	{
		unsigned char data[2];
		uint16_t val;
	} parm;
	socklen_t l;
	unsigned char *data;
	struct sockaddr_storage a;

	memset(&a,0,sizeof(a));

	if((a.ss_family=family)==AF_INET)
		memcpy(&((struct sockaddr_in *)&a)->sin_addr.s_addr,addr,4);
	else memcpy(((struct sockaddr_in6 *)&a)->sin6_addr.s6_addr,addr,16);

	if((s=socket(family,SOCK_RAW|SOCK_CLOEXEC|SOCK_NONBLOCK,
		IPPROTO_TCP))==-1)
	{
		perror("socket");
		return -1;
	}

	if(connect(s,(struct sockaddr *)&a,sizeof(a)))
	{
		perror("connect");
		close(s);
		return -1;
	}

	memset(&a,0,sizeof(a));
	l=sizeof(a);
	if(getsockname(s,(struct sockaddr *)&a,&l))
	{
		perror("getsockname");
		close(s);
		return -1;
	}

	if(family==AF_INET)
	{
		l=4;
		data=(unsigned char *)
			&((struct sockaddr_in *)&a)->sin_addr.s_addr;
	}
	else
	{
		l=16;
		data=(unsigned char *)
			((struct sockaddr_in6 *)&a)->sin6_addr.s6_addr;
	}

	*sum=0;

	parm.val=htons(sizeof(struct tcphdr));
	*sum+=parm.data[0];
	*sum+=((unsigned int)parm.data[1])<<8;

	parm.val=htons(IPPROTO_TCP);
	*sum+=parm.data[0];
	*sum+=((unsigned int)parm.data[1])<<8;

	for(i=0;i<l;i+=2)
	{
		*sum+=addr[i];
		*sum+=((unsigned int)addr[i+1])<<8;
		*sum+=data[i];
		*sum+=((unsigned int)data[i+1])<<8;
	}

	return s;
}

static int sendv4(int s,void *addr,int seq)
{
	int sum;
	int i;
	struct sockaddr_in a;
	union
	{
		struct icmphdr icmp;
		unsigned char data[64];
	} pkt;

	memset(&a,0,sizeof(a));
	memset(&pkt,0,sizeof(pkt));

	pkt.icmp.type=ICMP_ECHO;
	pkt.icmp.code=0;
	pkt.icmp.checksum=0;
	pkt.icmp.un.echo.id=htons(getpid());
	pkt.icmp.un.echo.sequence=htons(seq);

	for(sum=i=0;i<sizeof(pkt.data);)
	{
		sum+=pkt.data[i++];
		sum+=((int)(pkt.data[i++]))<<8;
	}
	while(sum&0xffff0000)sum=(sum>>16)+(sum&0xffff);
	sum=~sum;

	((unsigned char *)(&pkt.icmp.checksum))[0]=(unsigned char)sum;
	((unsigned char *)(&pkt.icmp.checksum))[1]=(unsigned char)(sum>>8);

	a.sin_family=AF_INET;
	a.sin_addr.s_addr=*((int *)addr);

	sendto(s,pkt.data,sizeof(pkt.data),0,(struct sockaddr *)&a,sizeof(a));

	return 0;
}

static int recvv4(int s,void *addr,int seq)
{
	socklen_t l;
	struct sockaddr_in a;
	struct
	{
		struct iphdr hdr;
		union
		{
			struct icmphdr icmp;
			unsigned char data[64];
		} icmp;
	} pkt;

	memset(&a,0,sizeof(a));
	memset(&pkt,0,sizeof(pkt));

	l=sizeof(a);
	if(recvfrom(s,&pkt,sizeof(pkt),0,(struct sockaddr *)&a,&l)
		!=sizeof(pkt))return -1;

	if(a.sin_family!=AF_INET||a.sin_addr.s_addr!=(*(int *)addr))return -1;
	if(pkt.icmp.icmp.type!=ICMP_ECHOREPLY||pkt.icmp.icmp.code)return -1;
	if(pkt.icmp.icmp.un.echo.id!=htons(getpid())||
		pkt.icmp.icmp.un.echo.sequence!=htons(seq))return -1;

	return 0;
}

static int sockv4(void)
{
	int s;
	struct icmp_filter f;

	if((s=socket(AF_INET,SOCK_RAW|SOCK_CLOEXEC|SOCK_NONBLOCK,
		IPPROTO_ICMP))==-1)
	{
		perror("socket");
		return -1;
	}

	f.data=~(1<<ICMP_ECHOREPLY);
	if(setsockopt(s,SOL_RAW,ICMP_FILTER,&f,sizeof(f)))
	{
		perror("setsockopt");
		close(s);
		return -1;
	}

	return s;
}

static int sendv6(int s,void *addr,int seq)
{
	int sum;
	int i;
	struct sockaddr_in6 a;
	union
	{
		struct icmp6_hdr icmp;
		unsigned char data[64];
	} pkt;

	memset(&a,0,sizeof(a));
	memset(&pkt,0,sizeof(pkt));

	pkt.icmp.icmp6_type=ICMP6_ECHO_REQUEST;
	pkt.icmp.icmp6_code=0;
	pkt.icmp.icmp6_cksum=0;
	pkt.icmp.icmp6_id=htons(getpid());
	pkt.icmp.icmp6_seq=htons(seq);

	for(sum=i=0;i<sizeof(pkt.data);)
	{
		sum+=pkt.data[i++];
		sum+=((int)(pkt.data[i++]))<<8;
	}
	while(sum&0xffff0000)sum=(sum>>16)+(sum&0xffff);
	sum=~sum;

	((unsigned char *)(&pkt.icmp.icmp6_cksum))[0]=(unsigned char)sum;
	((unsigned char *)(&pkt.icmp.icmp6_cksum))[1]=(unsigned char)(sum>>8);

	a.sin6_family=AF_INET6;
	memcpy(a.sin6_addr.s6_addr,addr,16);

	sendto(s,pkt.data,sizeof(pkt.data),0,(struct sockaddr *)&a,sizeof(a));

	return 0;
}

static int recvv6(int s,void *addr,int seq)
{
	socklen_t l;
	struct sockaddr_in6 a;
	union
	{
		struct icmp6_hdr icmp;
		unsigned char data[64];
	} pkt;

	memset(&a,0,sizeof(a));
	memset(&pkt,0,sizeof(pkt));

	l=sizeof(a);
	if(recvfrom(s,&pkt,sizeof(pkt),0,(struct sockaddr *)&a,&l)
		!=sizeof(pkt))return -1;

	if(a.sin6_family!=AF_INET6||memcmp(a.sin6_addr.s6_addr,addr,16))
		return -1;
	if(pkt.icmp.icmp6_type!=ICMP6_ECHO_REPLY||pkt.icmp.icmp6_code)return -1;
	if(pkt.icmp.icmp6_id!=htons(getpid())||pkt.icmp.icmp6_seq!=htons(seq))
		return -1;

	return 0;
}

static int sockv6(void)
{
	int s;
	struct icmp6_filter f;

	if((s=socket(AF_INET6,SOCK_RAW|SOCK_CLOEXEC|SOCK_NONBLOCK,
		IPPROTO_ICMPV6))==-1)
	{
		perror("socket");
		return -1;
	}

	ICMP6_FILTER_SETBLOCKALL(&f);
	ICMP6_FILTER_SETPASS(ICMP6_ECHO_REPLY,&f);
	if(setsockopt(s,IPPROTO_ICMPV6,ICMP6_FILTER,&f,sizeof(f)))
	{
		perror("setsockopt");
		close(s);
		return -1;
	}

	return s;
}

static struct
{
	int (*sck)(void);
	int (*snd)(int s,void *addr,int seq);
	int (*rcv)(int s,void *addr,int seq);
} rxtx[2]=
{
	{sockv4,sendv4,recvv4},
	{sockv6,sendv6,recvv6}
};

static void dummy(int unused)
{
}

static int icmpwait(void *addr,int delay,int max,int protocol)
{
	int proto=0;
	int r=-1;
	int seq=-1;
	struct pollfd p;
	struct sigevent ev;
	timer_t id;
	struct itimerspec tme;
	sigset_t set;

	if(protocol==AF_INET6)proto=1;

	if((p.fd=rxtx[proto].sck())==-1)goto err1;
	p.events=POLLIN;

	dropall();

	signal(SIGUSR1,dummy);

	sigemptyset(&set);
	sigaddset(&set,SIGUSR1);

	sigprocmask(SIG_BLOCK,&set,NULL);

	memset(&ev,0,sizeof(ev));
	ev.sigev_notify=SIGEV_SIGNAL;
	ev.sigev_signo=SIGUSR1;

	if(timer_create(CLOCK_MONOTONIC,&ev,&id))
	{
		perror("timer_create");
		goto err2;
	}

	tme.it_value.tv_sec=tme.it_interval.tv_sec=delay/1000;
	tme.it_value.tv_nsec=tme.it_interval.tv_nsec=(delay%1000)*1000000;

	if(timer_settime(id,0,&tme,NULL))
	{
		perror("timer_settime");
		goto err3;
	}

	rxtx[proto].snd(p.fd,addr,++seq);

	while(r)
	{
		sigprocmask(SIG_UNBLOCK,&set,NULL);

		switch(poll(&p,1,-1))
		{
		case -1:if(++seq==max&&max)goto err3;
			rxtx[proto].snd(p.fd,addr,seq);
		case 0:	sigprocmask(SIG_BLOCK,&set,NULL);
			continue;
		}

		sigprocmask(SIG_BLOCK,&set,NULL);

		if(!(p.revents&POLLIN))break;

		r=rxtx[proto].rcv(p.fd,addr,seq);
	}

err3:	timer_delete(id);
err2:	close(p.fd);
err1:	return r;
}

static int tcpwait(void *addr,int port,int delay,int max,int proto)
{
	int r=-1;
	int seq=-1;
	int srcsock;
	int srcport;
	unsigned int sum;
	struct pollfd p;
	struct sigevent ev;
	timer_t id;
	struct itimerspec tme;
	sigset_t set;

	if((p.fd=socktcp(proto,addr,&sum))==-1)goto err1;
	p.events=POLLIN;

	dropall();

	signal(SIGUSR1,dummy);

	sigemptyset(&set);
	sigaddset(&set,SIGUSR1);

	sigprocmask(SIG_BLOCK,&set,NULL);

	memset(&ev,0,sizeof(ev));
	ev.sigev_notify=SIGEV_SIGNAL;
	ev.sigev_signo=SIGUSR1;

	if(timer_create(CLOCK_MONOTONIC,&ev,&id))
	{
		perror("timer_create");
		goto err2;
	}

	tme.it_value.tv_sec=tme.it_interval.tv_sec=delay/1000;
	tme.it_value.tv_nsec=tme.it_interval.tv_nsec=(delay%1000)*1000000;

	if(timer_settime(id,0,&tme,NULL))
	{
		perror("timer_settime");
		goto err3;
	}

	if(sendtcp(p.fd,proto,port,&srcsock,&srcport,sum))
	{
		srcsock=-1;
		srcport=0;
	}

	while(r)
	{
		sigprocmask(SIG_UNBLOCK,&set,NULL);

		switch(poll(&p,1,-1))
		{
		case -1:if(srcsock!=-1)close(srcsock);
			if(++seq==max&&max)goto err3;
			if(sendtcp(p.fd,proto,port,&srcsock,&srcport,sum))
			{
				srcsock=-1;
				srcport=0;
			}
		case 0:	sigprocmask(SIG_BLOCK,&set,NULL);
			continue;
		}

		sigprocmask(SIG_BLOCK,&set,NULL);

		if(!(p.revents&POLLIN))break;

		r=recvtcp(p.fd,proto,port,srcport);
	}

	if(srcsock!=-1)
	{
		close(srcsock);
		srcsock=-1;
		srcport=0;
	}

err3:	timer_delete(id);
err2:	close(p.fd);
err1:	return r;
}

static int inet_xton(char *addr,void *dst,int *proto,int prefer)
{
	int r;
	struct addrinfo *ptr;
	struct addrinfo *list=NULL;
	struct addrinfo hints;

	switch(inet_pton(AF_INET,addr,dst))
	{
	case 1: *proto=AF_INET;
		return 0;

	case -1:perror("inet_pton");
		return -1;

	case 0: switch(inet_pton(AF_INET6,addr,dst))
		{
		case 1: *proto=AF_INET6;
			return 0;

		case -1:perror("inet_pton");
			return -1;
		}
		break;
	}

	memset(&hints,0,sizeof(hints));
	hints.ai_family=AF_UNSPEC;
	hints.ai_flags=AI_ADDRCONFIG|AI_IDN;

	r=getaddrinfo(addr,NULL,&hints,&list);

	if(!r&&list)
	{
		for(r=-1,ptr=list;ptr;ptr=ptr->ai_next)switch(ptr->ai_family)
		{
		case AF_INET:
			*((int *)dst)=((struct sockaddr_in *)ptr->ai_addr)
				->sin_addr.s_addr;
			*proto=AF_INET;
			r=0;
			if(prefer&1)goto out;

		case AF_INET6:
			memcpy(dst,((struct sockaddr_in6 *)
				ptr->ai_addr)->sin6_addr.s6_addr,16);
			*proto=AF_INET6;
			r=0;
			if(prefer&2)goto out;
		}

out:		freeaddrinfo(list);
	}
	else if(r)perror("getaddrinfo");
	else r=-1;

	return r;
}

static void usage(char *cmd)
{
	fprintf(stderr,
		"Usage: %s [<options>] address\n"
		"-4           prefer IPv4\n"
		"-6           prefer IPv6\n"
		"-p <port>    TCP ping to <port> instead of ICMP ping\n"
		"-w <msecs>   wait <msecs> for reply (default 1000)\n"
		"-c <count>   repeat <count> times (0=forever, default 0)\n",
		cmd);
	exit(1);
}

int main(int argc,char *argv[])
{
	int c;
	int proto;
	int port=0;
	int prefer=0;
	int delay=1000;
	int max=0;
	unsigned char addr[16];

	if(drop())
	{
		fprintf(stderr,"can't drop privileges, aborting\n");
		return 1;
	}

	while((c=getopt(argc,argv,"46w:c:p:"))!=-1)switch(c)
	{
	case '4':
		prefer|=1;
		break;

	case '6':
		prefer|=2;
		break;

	case 'w':
		delay=atoi(optarg);
		if(delay<100||delay>60000)usage(argv[0]);
		break;

	case 'c':
		max=atoi(optarg);
		if(max<0)usage(argv[0]);
		break;

	case 'p':
		port=atoi(optarg);
		if(port<1||port>65535)usage(argv[0]);
		break;

	default:usage(argv[0]);
	}

	if(optind!=argc-1)usage(argv[0]);

	if(!prefer)prefer=3;

	if(inet_xton(argv[optind],addr,&proto,prefer))
	{
		fprintf(stderr,"address lookup for %s failed\n",argv[optind]);
		return 1;
	}

	if(port)
	{
		if(tcpwait(addr,port,delay,max,proto))return 1;
	}
	else if(icmpwait(addr,delay,max,proto))return 1;

	return 0;
}
